﻿/*************************************************************************
 *	Simplified RTOS for Arduino Ver. 1.43 [Jun.05.2016] by T.Kohki
 *
 *	■簡易RTOSのTASK定義
 *	Flag[n]・・・状態遷移レジスタ
 *		bit 7 : Unloading状態にあり、タスク実行は行われません
 *		bit 1 : Ready Flag	trueの場合、タスク実行が可能です
 *		bit 0 : Waiting Flag trueの場合、Timerカウントアップ終了後にReadyに遷移します
 *
 *	Timer[n]・・・・・実行待ちカウンタ
 *		0 : カウント終了(Ready状態）
 *		t : カウント中(Waiting状態)
 *
 *	タスク実行周期（32bit）に関して
 *		0 : 自動起動しないタスク
 *		1 : 最少実行時間・・・毎回呼び出されます
 *		n : n サイクル後に実行されます
 *			実行時間精度はあまり高くありませんが、極端な誤差は出ません 
 *
 *	タスクの繰り返し実行について
 *		TimerSet[n] > 0 の場合、WaitingからReadyに切り替わる際、Waitingフラグが
 *		解除されません。
 *		この機能によって、指定されたタスクは、ほぼ一定の時間期間で再実行される形になりますので、
 *		ソフトウェアによるタイマー割り込み処理が実現されています。
 *		タスクの繰り返し実行を中止する場合は、End(n) で Flag[n]を 0 に切り替えます。
 *
 *	■ ライセンス条件
 *	  Released under the MIT license
 *	  http://opensource.org/licenses/mit-license.php 
 *
 *	  licenses/MIT_license - Open Source Group Japan Wiki
 *	  - Open Source Group Japan - SourceForge.JP
 *	  http://sourceforge.jp/projects/opensource/wiki/licenses%2FMIT_license
 *
 *************************************************************************/
#ifndef RTOS_H
#define RTOS_H

#include <stdint.h>

/*************************************************************************
 *	簡易RTOSのタスク実行に関する設定
 *	board_config.hの設定が優先されるので注意！
*************************************************************************/
#ifndef RTOS_MAXTASK
#define RTOS_MAXTASK	(8)						// 定時実行処理のタスク数 8n(8, 16, 24とする）
#endif

//	F_CPUが 1, 2, 4, 8, 16MHz以外のAVR版 Arduinoで正確な計時処理が必要な場合は、
//	RTOS_ADJUSTに補正値「(1000000L/F_CPU*64L*256L/2L)」を設定してください。
//	補正処理が必要ない場合は、定義しないか「1L」または「0L」を設定してください。
#ifndef RTOS_ADJUST
	//#define RTOS_ADJUST		(1000000L/F_CPU*64L*256L/2L)	// 計時分解能を考慮した調整値
	#define RTOS_ADJUST		(1L)			// 補正不要な場合
#endif

#define	RTOS_MINCOUNT	(500L)				//	マイコンの処理能力に応じて適宜決定する
											//	RX63 (96MHz) では、50usが下限限界値、200usが実用下限です。
#define RTOS_INVALID	(0xc0)					//	無効
#define RTOS_UNLOADED	(0x80)					//	未定義
#define RTOS_DORMANT	(0x00)					//	Dormant状態
#define RTOS_WAITING	(0x01)					//	Waiting状態
#define RTOS_READY		(0x02)					//	Ready状態
#define RTOS_CONTINUE	(0x04)					//	周期実行Flag

#define RTOS_EVENTDRIVEN	(0L)				//	定時実行されない時間設定
#define RTOS_DOEACHSCAN		(1L)				//	毎スキャン実行を指定

#define RTOS_CURRENTTASK	(0)					//	自タスク

//	RTOSの実行状態監視用設定
#define RTOS_INTERVAL	(30000000L)			//	30.0sec毎に最大スキャンタイムの計時記録を更新

/*************************************************************************
 *  RXマイコンのSysTickTimer.hで定義されているコード
 ************************************************************************/
//  RXマイコンでは、SysTick Counterが定義されていないため、システム時刻[us]を
//  正確に計測する場合は、CMTを２つ使って経過処理する必要があるため、AVRやARMマイコンと
//  処理が異なる。
//  [us]単位の時刻は、SYSTICK_US_COUNT（0x10000 = 65,536[ms]）毎にリセットされるため、
//  下記の計算式を使って求める。
//    PresentSysTick = micros();
//    Delta_us = (PresentSysTick - PreviousTime + SYSTICK_US_COUNT) % (SYSTICK_US_COUNT);
//    PreviousTime = PresentSysTick;
//  #define   SYSTICK_US_COUNT  (1000L*0x10000L)  //  [us] の桁上がりタイミング

/*************************************************************************
 *	空のタスク（ハングアップ防止用）
 ************************************************************************/
void	Rtos_Hook(void)	__attribute__((weak));;	//	空のタスク（ハングアップ防止用）
//void	Rtos_Hook(void);							//	CS+での表現

/*************************************************************************
 *  Class definition of RTOS
 *************************************************************************/
#ifdef __cplusplus
class RTOS
{
public:
	//	Default Constructor
	RTOS(void);
	//	Class Member
	uint8_t	Loading(uint8_t, uint32_t us, void (*f)());	//	タスクの登録
	uint8_t	Unloading(uint8_t);							//	タスクの登録解除
	uint8_t	Timer(uint8_t TaskId, uint32_t us);			//	Change Timer Setting of Task
	uint8_t	IdNum(void);								//	現在実行中のタスク番号を返す
	
	void 	Begin(uint8_t TaskId);						//	タスクの起動(Readyに遷移)
	void 	End(uint8_t TaskId);              //	タスクの休止(Dormantに遷移)
	void 	Sleep(uint8_t TaskId, uint32_t TimeOut);	//	タスクの一定時間休止(Waitingに遷移)
	void 	Suspend(uint8_t TaskId);					//	タスクの休止（ReadyからWaitingに遷移）
	void 	Resume(uint8_t TaskId);						//	タスクの再開（DormantからWaitingに遷移）
  void  Continue(uint8_t TaskId);         //  タスクの繰り返し実行（Continue Flagをセット）
	int8_t 	status(uint8_t TaskId);         //	Task Statusの参照

	uint32_t AdjustCount(uint32_t us);					//	タイマー設定値の補正処理

	void		Control(void);							//	タスクディスパッチ制御

	uint32_t	ScanTimeMax(void);						//	スキャンタイム最大値
	uint32_t	ScanTimeAvg(void);						//	スキャンタイム平均値

// -----------------------------------------------------------------------
private:
	// 共有変数
	volatile int8_t			NumOfTask;					//	登録されたTaskの数
	volatile uint8_t			Id;							//	現在実行中のタスク番号

	//	静的変数定義（インスタンスは１つ）
	volatile int32_t		TimerSet[RTOS_MAXTASK];			// タスクの実行周期設定[us]
	volatile int32_t		Counter[RTOS_MAXTASK];			// タスク実行までの待ち時間[us]
	volatile uint8_t		Flag[RTOS_MAXTASK];				// タスクの状態遷移
	void		(*Task[RTOS_MAXTASK])();		//	関数ポインタ

	volatile int32_t		Delta_us;						//	Differential Time[us]
	volatile int32_t		PresentSysTick;					//	現在の時刻
	volatile int32_t		PreviousTime;					//	１回前の時刻

	//	実行状況の確認
	volatile int32_t	scan_max;							//	5秒更新
	volatile int32_t	scan_max2;							//	5秒更新用のワーク変数
	volatile int32_t	scan_avg;							//	簡易LPF処理結果
	volatile int32_t	interval;							//	5秒計時用変数
};

/**************************************************************************
 *	グローバル変数定義
 **************************************************************************/
extern RTOS Rtos;
#endif	//	class RTOS
#endif	//	RTOS_h

/**************************************************************************
 *	■来歴管理
 *  Ver. 1.43 [Jun.05.2016] by T.Kohki
 *  1,  Arduinoで動作するように書き換え（IDEでコンパイルできるように修正)
 *  2,  Continue命令を追加
 *  3,  MITライセンスで一般公開
 *  
 *	Ver. 1.40 [Jan.27.2016] by T.Kohki
 *	1,	RX63 マイコン用に書き換え（CS+でコンパイルできるように改版)
 *	2,	CMT0とCMT1を使ったSystickTimer.cppとの併用で、正確な[us]計時が行えます。
 *
 *	Ver. 1.32 [Jul.26.2015] by T.Kohki
 *	1,	Suspend処理はReady Flagをクリアする形に修正
 *		なぜか、計時処理が安定するようになった。
 *
 *	Ver. 1.31 [Jul.12.2015] by T.Kohki
 *	1,	LPC1114FN28版は正確な1ms SysTickTimer割り込み処理を使うため、micors()の
 *		計時処理補正が不要になったため、補正処理ソフトを変更
 *	2,  計時補正処理は、F_CPUが2, 4, 8, 16MHz以外のAVRマイコンで行う
 *		ただし、計時補正処理はコードサイズ、実行時間共に長いため、補正不要指定できるように変更
 *
 *	Ver. 1.30 [Jun.27.2015] by T.Kohki
 *	1,	最小実行時間と計時処理の方法を見直し
 *	2,	F_CPUが8MHz, 16MHz以外だとmicros()関数で丸め誤差が発生するため、設定値の補正を追加した
 *	3,	周期実行フラグを新たに定義
 *
 *	Ver. 1.20 [May.23.2015] by T.Kohki
 *	1,	状態遷移フラグの管理方法、コマンド名称を見直し
 *
 *	Ver. 1.11 [May.09.2015] by T.Kohki
 *	1,	タスクの呼び出し方法を１スキャンで全タスク実行から１スキャン１タスクに変更
 *	2,	Arduinoのライブラリとして利用できるような形に変更
 *	3,	Task unloadに対応
 *	4,	タイマーカウントの補正用として計時調整値(RTOS_ADJUST)を採用. 8MHzで130us程度
 *	5,	8MHzにて最小スキャンタイム at 1 taskは、約1ms
 *
 *	Ver. 1.10 [May.06.2015] by T.Kohki
 *	1, Arduino用初期版としてC++に移植、基本動作を確認した
 *
 *	Ver. 1.00 [May.06.2013] by T.Kohki
 *	1, AVR StudioのC言語版
 *
 **************************************************************************/
